import datetime
import decimal
import enum
import json
import pathlib
import re
import uuid
from dataclasses import dataclass
from types import SimpleNamespace
from typing import Any

from project.settings import BASE_DIR


def get_operations(query):
    Substring = "(query \w+|mutation \w+)"
    operations = [
        f.replace("query", "").replace("mutation", "").strip()
        for f in re.findall(Substring, query, re.IGNORECASE)
    ]
    return sorted(operations)


def role_operations():
    all_roles = {}
    root = BASE_DIR / "roles"
    for path in pathlib.Path(root).glob("*"):
        role_query = ""
        if path.is_dir():
            for file in path.glob("*.graphql"):
                with open(file, "r") as f:
                    role_query += "\n"
                    role_query += f.read()
            all_roles[path.name] = role_query
    return all_roles


class Roles:
    def __init__(self, schema=None):
        # Roles
        roles = role_operations()

        # Operations
        __operations__ = list(
            set(get_operations("\n".join([f for f in roles.values()])))
        )
        operations = enum.Enum("Operations", {k: k for k in __operations__})

        # Init: Setup
        self.schema = schema
        self.roles = roles
        self.all_operations = operations
        self._all_operations = __operations__

    def operations(self, role=None):
        query = self.roles.get(role)
        if query:
            ops = get_operations(query)
        else:
            ops = None
        return ops

    def execute_sync(self, role=None, operation=None, variables=None):
        query = self.roles.get(role)
        if operation in query and query:
            resp = self.schema.execute_sync(
                query,
                operation_name=operation,
                variable_values=variables,
            )
            return dict(data=resp.data, errors=resp.errors)
        return None

    async def execute(self, role=None, operation=None, variables=None):
        query = self.roles.get(role)
        if query and operation in query:
            resp = await self.schema.execute(
                query,
                operation_name=operation,
                variable_values=variables,
            )
            return dict(data=resp.data, errors=resp.errors)
        return None


@dataclass
class Instance:
    value: Any = None
    default: Any = None
    required: bool = False
    isObject: bool = False
    type: str = None


class NestedNamespace(SimpleNamespace):
    @staticmethod
    def map_entry(entry):
        if isinstance(entry, dict):
            return NestedNamespace(**entry)

        return entry

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for key, val in kwargs.items():
            if type(val) == dict:
                setattr(self, key, NestedNamespace(**val))
            elif type(val) == list:
                setattr(self, key, list(map(self.map_entry, val)))


def try_json(value):
    try:
        return json.loads(value)
    except:
        return value


SCALAR_TYPES = [
    "String",
    "Int",
    "Float",
    "Boolean",
    "ID",
    "Date",
    "DateTime",
    "Time",
    "Decimal",
    "UUID",
    "Void",
]


def get_scalar_type(value):
    if isinstance(value, int):
        return "Int"
    elif isinstance(value, str):
        return "String"
    elif isinstance(value, float):
        return "Float"
    elif isinstance(value, bool):
        return "Boolean"
    elif isinstance(value, decimal.Decimal):
        return "Decimal"
    elif isinstance(value, datetime.date):
        return "Date"
    elif isinstance(value, datetime.datetime):
        return "DateTime"
    elif isinstance(value, datetime.time):
        return "Time"
    elif isinstance(value, uuid.UUID):
        return "UUID"
    elif value == None:
        return "Void"
    else:
        return "Unknown"


def create_gql_core_setup(list_to_check: list = []):
    queries = {}
    for query in list_to_check:
        q = NestedNamespace(**query)
        queries[q.name] = {}
        for field in q.args:
            setup = Instance(
                default=None
                if field.defaultValue == "null"
                else try_json(field.defaultValue),
                required=(field.type.kind == "NON_NULL"),
                isObject=False,
            )
            if field.type.ofType:
                type_name = field.type.ofType.name
                setup.type = type_name
                inputFields = field.type.ofType.inputFields
                is_object = field.type.ofType.kind == "INPUT_OBJECT"
                setup.isObject = is_object
                if inputFields:
                    items = {}
                    for inputKey in inputFields:
                        defaultValue = try_json(inputKey.defaultValue)
                        type_name = inputKey.type.name
                        if not type_name:
                            type_name = get_scalar_type(defaultValue)
                        setup_fields = Instance(default=defaultValue, type=type_name)
                        items[inputKey.name] = setup_fields.__dict__
                    setup.default = items
            if (
                not setup.default
                and not setup.required
                and not setup.required
                and not setup.type
            ):
                setup.type = "String"
            queries[q.name][field.name] = setup.__dict__
    return queries


def make_gql_query(query_list: list = [], is_mutation: bool = False):
    query_items = []
    for query, fields in query_list.items():
        if is_mutation:
            query_text = f"""mutation {query}"""
        else:
            query_text = f"""query {query}"""
        input_setup = []
        query_setup = []
        is_page = False
        for name, setup in fields.items():
            if not setup.get("isObject"):  # Scalar
                type_name = setup.get("type")
                required = "!" if setup.get("required") else ""
                type_name += required
                if type_name:
                    input_setup.append(f"${name}: {type_name}")
                    query_setup.append(f"{name}: ${name}")

            else:
                input_setup_c = []
                query_setup_c = {name: []}
                for name_c, setup_c in setup.get("default", {}).items():
                    if name.lower() == "pagination":
                        is_page = True
                    type_name = setup_c.get("type")
                    required = "!" if setup_c.get("required") else ""
                    type_name += required
                    if type_name:
                        input_setup_c.append(f"${name_c}: {type_name}")
                        query_setup_c[name].append(f"{name_c}: ${name_c}")
                input_setup.extend(input_setup_c)
                query_setup.append(
                    f"""{name}: {{ {", ".join(query_setup_c[name])} }}"""
                )
        if is_mutation:
            query_output = """
    ... on Error {
        error
        type
        messages
        meta
    }""".strip()
        elif not is_page:
            query_output = "id"
        else:
            query_output = """
    pageInfo {
      pages
    }
    edges {
      node {
        id
      }
    }
    """.strip()

        query_string = f"""{query_text} ({", ".join(input_setup)}) {{\n {query} ({", ".join(query_setup)}) {{ \n {query_output} \n }} \n}}"""
        query_items.append(query_string)
    return "\n\n".join(query_items)


def create_js_forms(all_forms: list = []):
    output = {}
    for schema, fields in all_forms.items():
        output[schema] = {}
        for field, setup in fields.items():
            is_object = setup.get("isObject")
            default_value = setup.get("default")
            if is_object:
                output[schema][field] = {}
                for fkey, fsetup in default_value.items():
                    output[schema][field][fkey] = fsetup.get("default")
            else:
                output[schema][field] = default_value
    return output


def make_gql_setup_core(query: list = [], mutation: list = []):
    setup = {}
    if query:
        setup["query"] = create_gql_core_setup(query)
    if mutation:
        setup["mutation"] = create_gql_core_setup(mutation)
    all_forms = {}
    all_forms.update(setup.get("query", {}))
    all_forms.update(setup.get("mutation", {}))
    actions_example = (
        make_gql_query(setup.get("query", {}))
        + "\n\n"
        + make_gql_query(setup.get("mutation", {}), True)
    )
    return_value = {
        "keys": sorted(list(all_forms.keys())),
        "schema": all_forms,
        "form": create_js_forms(all_forms),
        "code": actions_example,
    }
    return return_value


def js_forms_code(forms):
    code = f"""const FORMS = { forms };"""
    code += "\n\n"
    code += """
class GqlForm {
  constructor(setup) {
    this.$default = () => ({ ...setup });
    this.reset();
  }

  static $create(setup) {
    return new GqlForm(setup);
  }

  reset() {
    this.form = this.$default();
  }
}

function gqlSetup(setup) {
  const output = {};
  Object.keys(setup).forEach((form) => {
    const model = GqlForm.$create(setup[form]);
    output[form] = model;
  });
  return output;
}

const SETUP = gqlSetup(FORMS);
SETUP.$keys = Object.keys(FORMS)

export default SETUP;
""".strip()

    return code


INFO_QUERY = """
fragment InfoIntrospection on __Type {
  return: fields {
    name
    args {
      name
      defaultValue
      type {
        kind
        ofType {
          name
          kind
          inputFields {
            name
            defaultValue
            type {
              name
            }
          }
        }
      }
    }
  }
}

query availableQueries {
  schema: __type(name: "Query") {
    ...InfoIntrospection
  }
}

query availableMutations {
  schema: __type(name: "Mutation") {
    ...InfoIntrospection
  }
}
"""


def build_gql_setup_files(setup):
    roles_dir = BASE_DIR / "roles"
    static_dir = BASE_DIR / "static"

    if setup.get("code"):
        code_text = setup.get("code")
        del setup["code"]

    if setup.get("all_operations"):
        all_operations = setup.get("all_operations")
        del setup["all_operations"]

    with open(str(roles_dir / "actions.graphql"), "w") as file:
        file.write(code_text)
    with open(str(static_dir / "schema.js"), "w") as file:
        file.write(f"""export default { json.dumps(setup, indent=4) }""")
    with open(str(static_dir / "forms.js"), "w") as file:
        code = js_forms_code(json.dumps(setup.get("form"), indent=4))
        file.write(code)
    with open(str(static_dir / "operations.js"), "w") as file:
        file.write(f"""export default { json.dumps(all_operations, indent=4) }""")
    return setup


def gql_schema_info(schema):
    # Core Setup
    queries_resp = schema.execute_sync(INFO_QUERY, operation_name="availableQueries")
    mutations_resp = schema.execute_sync(
        INFO_QUERY, operation_name="availableMutations"
    )
    query = queries_resp.data["schema"]["return"]
    mutation = mutations_resp.data["schema"]["return"]
    setup = make_gql_setup_core(query=query, mutation=mutation)
    # Add Roles
    Role = Roles(schema)
    setup["all_operations"] = Role._all_operations
    build_gql_setup_files(setup)
    return setup



import dataclasses
from types import SimpleNamespace

import strawberry
from graphql import GraphQLError
from strawberry.extensions import Extension
from strawberry.types import Info

from routers.users import decode_token, get_user


@dataclasses.dataclass
class Method:
    name: str = None
    field: str = None
    action: str = "read"


def get_token_bearer(info):
    token_bearer = info.context["request"].headers.get("AUTHORIZATION")
    token = None
    if token_bearer:
        token = token_bearer.replace("Bearer ", "", 1)
    return token


async def get_token(context):
    token = get_token_bearer(context)
    user = None
    if token:
        user_token = await decode_token(token)
        username = user_token.get("username")
        if username:
            user = get_user(username)
    return user


def find_field_type(info):
    output = None
    name = info.parent_type.name
    if name == "Mutation":
        output = Method(action="write", name=info.field_name)
    elif (
        name
        not in [
            "Query",
            "PageInfo",
        ]
        and not name.endswith("Edge")
        and not name.endswith("Connection")
    ):
        output = Method(action="read", name=name.lower(), field=info.field_name)
    return output


AUTHORIZED_EXCEPTION = GraphQLError("You Are Not Authorized.")


class LoginRequired(Extension):
    async def on_executing_start(self):
        user = await get_token(self.execution_context)
        self.execution_context.context["user"] = user
        if not user:
            self.execution_context.result = SimpleNamespace(
                data=None,
                errors=[AUTHORIZED_EXCEPTION],
            )



from .db import Mongo, Sql
from .imports import settings

try:
    DATABASES = settings.DATABASES
except:
    DATABASES = None

if DATABASES:
    # SETUP
    mongo_setup = DATABASES.get("MONGO", {})
    sql_setup = DATABASES.get("SQL", {})

    # URL
    mongo_url = mongo_setup.get("URL")
    sql_url = sql_setup.get("URL")

    # DATABASE-NAME
    mongo_db = mongo_setup.get("NAME")

    # MANAGERS
    sql = Sql(url=sql_url)
    mongo = Mongo(url=mongo_url, name=mongo_db)

else:
    # MANAGERS
    sql = None
    mongo = None



from types import SimpleNamespace

# SQL-Manager
from databases import Database

# SQL-Controller
from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base

from dbcontroller import Sql

# Base
Base = declarative_base()


def init_sql(url: str = None):
    if url:
        # Setup
        engine = create_engine(url, echo=True)

        # Create-Tables
        def create_tables():
            Base.metadata.create_all(engine)
            return Base.metadata.tables

        database = Database(url)
        return SimpleNamespace(
            manager=Sql(database),
            base=Base,
            create_tables=create_tables,
            database=database,
        )
    return None



from dataclasses import field

import strawberry

from .pagination import Connection
from .typing import Dict, List


@strawberry.type
class Error:
    type: str = None
    messages: List = field(default_factory=list)
    meta: Dict = field(default_factory=dict)
    error: bool = True


def Query(Model):
    return Model | None


def Mutation(Model):
    return Model | Error | None


def Edges(Model):
    return Connection[Model] | None



from dbcontroller.manager.mongo import MongoCrud
from dbcontroller.manager.sql import ManagerCrud

from .plugins import get_fields

get_models = lambda installed_apps: [
    app for app in installed_apps.keys() if app.endswith(".models")
]
get_api = lambda installed_apps: [
    app for app in installed_apps.keys() if app.endswith(".api")
]


is_sql = lambda model: isinstance(model, ManagerCrud)
is_mongo = lambda model: isinstance(model, MongoCrud)


def model_to_dict(installed_apps):
    all_models = dict()
    for model_module in get_models(installed_apps):
        module = installed_apps.get(model_module)
        fields = get_fields(module)
        for field in fields:
            model = getattr(module, field)
            if is_sql(model):
                manager = model
                all_models[field.lower()] = manager
            elif is_mongo(model):
                all_models[field.lower()] = model
    return all_models



import json
from typing import Any, NewType

import strawberry


def load_list(v):
    val = json.loads(v)
    return val if isinstance(val, list) else []


def load_dict(v):
    val = json.loads(v)
    return val if isinstance(val, dict) else {}


JSONScalar = strawberry.scalar(
    NewType("JSON", Any),
    serialize=lambda v: v,
    parse_value=lambda v: json.loads(v),
    description="The JSON scalar type represents a generic GraphQL scalar value that could be: <Array or Object>.",
)

ListScalar = strawberry.scalar(
    NewType("List", Any),
    serialize=lambda v: v,
    parse_value=lambda v: load_list(v),
    description="The List scalar type represents a generic GraphQL scalar value that could be: <List or Array>.",
)

DictScalar = strawberry.scalar(
    NewType("Dict", Any),
    serialize=lambda v: v,
    parse_value=lambda v: load_dict(v),
    description="The Dict scalar type represents a generic GraphQL scalar value that could be: <Dict or Object>.",
)



from dataclasses import field
from typing import Generic, TypeVar

import strawberry

from fastberry.graphql.typing import Dict

GenericType = TypeVar("GenericType")


@strawberry.type
class Connection(Generic[GenericType]):
    """Represents a paginated relationship between two entities"""

    page_info: "PageInfo"
    edges: list["Edge[GenericType]"]


@strawberry.type
class PageInfo:
    """Pagination context to navigate objects with cursor-based pagination"""

    length: int
    pages: int
    extra: Dict = field(default_factory=dict)


@strawberry.type
class Edge(Generic[GenericType]):
    """An edge may contain additional information of the relationship. This is the trivial case"""

    node: GenericType
    cursor: str


class Response:
    def __init__(self):
        self.__edges = []

    def add(self, item):
        edge = Edge(node=item, cursor=item.id)
        self.__edges.append(edge)

    def __call__(self, length: int = None, pages: int = None, extra: dict = {}):
        edges = self.__edges
        if len(edges) > 0:
            return Connection(
                page_info=PageInfo(
                    length=length,
                    pages=pages,
                    extra=extra,
                ),
                edges=edges,
            )
        return None



import importlib
import pkgutil

get_fields = lambda model: [i for i in dir(model) if not i.startswith("__")]


def iter_namespace(ns_pkg):
    """https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/"""
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def import_modules(all_apps):
    """https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/"""
    installed_apps = {}
    for app in all_apps:
        try:
            module = importlib.import_module(app)
        except:
            module = None
        if module:
            discovered_plugins = {
                name: importlib.import_module(name)
                for finder, name, ispkg in iter_namespace(module)
            }
            installed_apps.update(discovered_plugins)
    return installed_apps



# Core
from .database import mongo, sql
from .imports import settings

# Loop Over Apps
from .models import model_to_dict
from .plugins import import_modules

"""
########  ########   #######        ## ########  ######  ######## 
##     ## ##     ## ##     ##       ## ##       ##    ##    ##    
##     ## ##     ## ##     ##       ## ##       ##          ##    
########  ########  ##     ##       ## ######   ##          ##    
##        ##   ##   ##     ## ##    ## ##       ##          ##    
##        ##    ##  ##     ## ##    ## ##       ##    ##    ##    
##        ##     ##  #######   ######  ########  ######     ##    
"""

ALL_PLUGINS = {}


def init_models():
    global ALL_PLUGINS
    all_models = None
    if settings and hasattr(settings, "INSTALLED_APPS"):
        installed_apps = import_modules(settings.INSTALLED_APPS)
        all_models = model_to_dict(installed_apps)
    if all_models:
        ALL_PLUGINS.update(all_models)
    return all_models


class Model:
    __plugins = ALL_PLUGINS

    @classmethod
    def get(cls, model):
        return cls.__plugins.get(model)

    @classmethod
    def all(cls):
        return cls.__plugins

    @classmethod
    def keys(cls):
        return list(cls.__plugins.keys())


# print(all_models.keys())
# print(installed_apps.keys())



from types import SimpleNamespace

# Mongo-Controller
import motor.motor_asyncio

# Mongo-Manager
from dbcontroller import Mongo


def init_mongo(url: str = None, name: str = None):
    if url and name:
        # Setup
        MONGO_DB = motor.motor_asyncio.AsyncIOMotorClient(url)
        # Mongo
        MONGO_DB[name]
        return SimpleNamespace(manager=Mongo(MONGO_DB)(name), base=MONGO_DB[name])
    return None



import os
import sys
from pathlib import Path

try:
    # Get "main.py" <FILE>
    MAIN = sys.modules.get("main")
    if MAIN:
        BASE_DIR = Path(MAIN.__file__).resolve().parent
    else:
        BASE_DIR = Path(sys.argv[0]).resolve().parent

    print(BASE_DIR)
    # Append "PROJECT" <Folder>
    sys.path.insert(0, os.path.join(BASE_DIR, "project"))

    # Import settings.py <FILE>
    from project import settings

    # Get "INSTALLED-APPS"
    if hasattr(settings, "INSTALLED_APPS"):
        INSTALLED_APPS = settings.INSTALLED_APPS
    else:
        INSTALLED_APPS = []

    # Include apps/
    sys.path.insert(0, os.path.join(BASE_DIR, "apps"))
except:
    raise Exception("Please create the <folder> ./project and the settings.py <file>.")
